#include <pthread.h>
#include <ncurses.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/select.h>

#include "common.h"
#include "protocol.h"

// General stuff
int request_socket;
int numberOfConnections = 0;
struct node *firstNode = NULL;

// Select stuff
pthread_t thread_select;
pthread_mutex_t con_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t sel_mutex = PTHREAD_MUTEX_INITIALIZER;
fd_set rfds;
int highestFd = 0;

struct connection {
	struct sockaddr_in clientaddr;
	int socket;
};

struct node {
	struct connection *con;
	struct node *next;
};

void addConnection(int socket, struct sockaddr_in *clientaddr);
void removeConnection(struct node *node);
void * runSelectLoop();
void broadcast(packet_t *packet);
void freeNode(struct node *node);

// Usage: ./server
int main(int argc, char *argv[]) {
	struct sockaddr_in serveraddr;
	struct sockaddr_in clientaddr;
	int clientaddrlen;
	int test;
	
	// Listen for connections
	printf("Entering server mode! \n");
	
	// Sockets identify a connection
	request_socket = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	
	if (request_socket < 0) {
		perror("Could not create socket!\n");
		exit(1);
	}

	int yes = 1;
	// Avoid "address already in use" error
	if (setsockopt(request_socket, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1) {
      	perror("setsockopt:");
      	return -1;
    }
    
    bzero(&serveraddr, sizeof(serveraddr));
    serveraddr.sin_family = AF_INET;
    serveraddr.sin_port = htons(PORT);
    serveraddr.sin_addr.s_addr = INADDR_ANY;
    
    // Bind serveraddress to our socket
    test = bind(request_socket, (struct sockaddr*)&serveraddr, sizeof(serveraddr));
    if (test < 0) {
    	perror("Failed to bind socket!\n");
    	exit(1);
    }
    
    test = listen(request_socket, SOMAXCONN);
    
    if (test < 0) {
    	perror("Failed to initialize listen queue!\n");
    	exit(1);
    }
    
    printf("Waiting for connections ...\n");
    
    clientaddrlen = sizeof(struct sockaddr_in);
    
    while (1) {
	    // Klientens adresse legges inn i clientaddr
	    int sock = accept(request_socket, (struct sockaddr*)&clientaddr, &clientaddrlen);

		if (numberOfConnections == 0) {
			// Create select thread
			
			// Runs the thread specified by argument 0
			// The thread that is created immediately calls the functions
			// specified by argument 2
			test = pthread_create(&thread_select, NULL, &runSelectLoop, NULL);
			if (test != 0) 
				perror("ERROR pthread");
		}
		pthread_mutex_lock(&con_mutex);	    
	    addConnection(sock, &clientaddr);
	    pthread_mutex_unlock(&con_mutex);
	    
	}
}

// Add a connection to the linked list
// Assumes the con_mutex is locked
void addConnection(int socket, struct sockaddr_in *clientaddr) {
	struct node *current;
	
	// Create the connection struct
	struct connection *newCon = malloc(sizeof(struct connection));
	newCon->socket = socket;
	memcpy(&newCon->clientaddr, clientaddr, sizeof(struct sockaddr_in));
	
	// Create the node
	struct node *newNode = malloc(sizeof(struct node));
	newNode->con = newCon;
	newNode->next = NULL;
		
	if (numberOfConnections == 0) {
		firstNode = newNode;
	} else {
		// Find last node of the list
		current = firstNode;
		while (current->next != NULL)
			current = current->next;
			
		// Attack the new node at the end of the list
		current->next = newNode;
	}
	
	// Add the new socket to the file descriptor set!
	highestFd = max(highestFd, socket);
	FD_SET(socket, &rfds);
	printf("Adding FD: %d\n", socket);
	
	numberOfConnections++;
	printf("New connection added to the conversation!\n");
}

// Entrypoint for the select thread
void * runSelectLoop() {
	struct node *current;
	struct timeval timeout;
	int socket;
	int test;
	char buf[PACKSIZE];
	packet_t *packetPtr;
	fd_set tmpFd;

	timeout.tv_sec = 0;
	timeout.tv_usec = 50000;

	while(1) {
		tmpFd = rfds;
		test = select(highestFd + 1, &tmpFd, NULL, NULL, &timeout);
		
		// Error
		if (test == -1) {
			perror("ERROR on select()");
		}
		
		// We've received something on one of our sockets
		if (test > 0) {
			// Go through all the connections
			pthread_mutex_lock(&con_mutex);
			current = firstNode;
		
			while (current != NULL) {
				// If we've received data on a socket, read the data and broadcast
				if (FD_ISSET(current->con->socket, &tmpFd)) {
					recv(current->con->socket, buf, PACKSIZE, MSG_DONTWAIT);
					packetPtr = (packet_t *) buf;
					
					if (strcmp(packetPtr->username, "ENDCON") == 0) {
						// Remove this socket from the FD set
						FD_CLR(current->con->socket, &rfds);
						
						// Remove this connection from the linked list
						struct node *tmp = current;
						current = current->next;
						removeConnection(tmp);
						
						// Continue to avoid moving current two times.
						continue;
					} else {
						broadcast(packetPtr);
					}
				}
				current = current->next;
			}
			pthread_mutex_unlock(&con_mutex);
		}
	}
	return NULL;
}

// Removes a connection from the linked list
// Assumes the con_mutex is locked
void removeConnection(struct node *node) {
	struct node *current;
	struct node *tmp;

	if (node == firstNode)
		firstNode = firstNode->next;
	else {
		current = firstNode;
		while (current->next != node) {
			current = current->next;
		}
		tmp = current->next;
		current->next = current->next->next;
		freeNode(tmp);
	}
}

// Broadcast a packet to all clients
// Assumes the con_mutex is locked
void broadcast(packet_t *packet) {
	struct node *current;

	current = firstNode;
	while (current != NULL) {
		write(current->con->socket, packet, sizeof(packet_t));
		current = current->next;
	}
}

// Free all memory allocated by a node
void freeNode(struct node *node) {
	free(node->con);
	free(node);
}
